module Equality = struct

(* Version of Meson.create_equality_axioms with added symmetry rule *)
let create_equality_axioms =
  let eq_thms = (CONJUNCTS o prove)
   (`(x:A = x) /\ (~(x = y) \/ y = x) /\
     (~(x:A = y) \/ ~(x = z) \/ (y = z))`,
    REWRITE_TAC[] THEN ASM_CASES_TAC `x:A = y` THEN
    ASM_REWRITE_TAC[] THEN CONV_TAC TAUT) in
  let imp_elim_CONV = REWR_CONV
    (TAUT `(a ==> b) <=> ~a \/ b`) in
  let eq_elim_RULE =
    MATCH_MP(TAUT `(a <=> b) ==> b \/ ~a`) in
  let veq_tm = rator(rator(concl(hd eq_thms))) in
  let create_equivalence_axioms (eq,_) =
    let tyins = type_match (type_of veq_tm) (type_of eq) [] in
    map (INST_TYPE tyins) eq_thms in
  let rec tm_consts tm acc =
    let fn,args = strip_comb tm in
    if args = [] then acc
    else itlist tm_consts args (insert (fn,length args) acc) in
  let rec fm_consts tm ((preds,funs) as acc) =
    try fm_consts(snd(dest_forall tm)) acc with Failure _ ->
    try fm_consts(snd(dest_exists tm)) acc with Failure _ ->
    try let l,r = dest_conj tm in fm_consts l (fm_consts r acc)
    with Failure _ -> try
        let l,r = dest_disj tm in fm_consts l (fm_consts r acc)
    with Failure _ -> try
        let l,r = dest_imp tm in fm_consts l (fm_consts r acc)
    with Failure _ -> try
         fm_consts (dest_neg tm) acc with Failure _ ->
    try let l,r = dest_eq tm in
        if type_of l = bool_ty
        then fm_consts r (fm_consts l acc)
        else failwith "atomic equality"
    with Failure _ ->
    let pred,args = strip_comb tm in
    if args = [] then acc else
    insert (pred,length args) preds,itlist tm_consts args funs in
  let create_congruence_axiom pflag (tm,len) =
    let atys,rty = splitlist (fun ty -> let op,l = dest_type ty in
                                        if op = "fun" then hd l,hd(tl l)
                                        else fail())
                             (type_of tm) in
    let ctys = fst(chop_list len atys) in
    let largs = map genvar ctys
    and rargs = map genvar ctys in
    let th1 = rev_itlist (C (curry MK_COMB)) (map (ASSUME o mk_eq)
        (zip largs rargs)) (REFL tm) in
    let th2 = if pflag then eq_elim_RULE th1 else th1 in
    itlist (fun e th -> CONV_RULE imp_elim_CONV (DISCH e th)) (hyp th2) th2 in
  fun tms -> let preds,funs = itlist fm_consts tms ([],[]) in
             let eqs0,noneqs = partition
                (fun (t,_) -> is_const t && fst(dest_const t) = "=") preds in
             if eqs0 = [] then [] else
             let pcongs = map (create_congruence_axiom true) noneqs
             and fcongs = map (create_congruence_axiom false) funs in
             let preds1,_ =
               itlist fm_consts (map concl (pcongs @ fcongs)) ([],[]) in
             let eqs1 = filter
               (fun (t,_) -> is_const t && fst(dest_const t) = "=") preds1 in
             let eqs = union eqs0 eqs1 in
             let equivs =
               itlist (union' equals_thm o create_equivalence_axioms)
                      eqs [] in
             equivs@pcongs@fcongs

let create_eq_axs sym = match sym with
    true -> create_equality_axioms
  | false -> Meson.create_equality_axioms

end
