(*
  Some examples that are not provable by MESON or METIS in reasonable time:
   LEANCOP [EXTENSION; IN_INSERT] (concl INSERT_COMM);;
   LEANCOP [EXTENSION; IN_DELETE] (concl DELETE_COMM);;
   LEANCOP [EXTENSION; IN_DELETE; IN_INTER] (concl DELETE_INTER);;
   LEANCOP [SUM_SUPERSET; SUBSET; IN_UNION] (concl SUM_UNION_RZERO);;
   LEANCOP [SUBSET; NSUM_SUPERSET; IN_UNION] (concl NSUM_UNION_RZERO);;
   LEANCOP [EXTENSION; NOT_IN_EMPTY; IN_INTERS; IN_INSERT] (concl INTERS_1);;
 *)

needs "cop/utils.ml";;
needs "cop/strom.ml";;
needs "cop/term.ml";;
needs "cop/print.ml";;
needs "cop/mapping.ml";;
needs "cop/database.ml";;
needs "cop/substitution.ml";;
needs "cop/cop.ml";;
needs "cop/equality.ml";;
needs "cop/prepare.ml";;
needs "cop/proof.ml";;
needs "cop/search.ml";;
needs "cop/recon.ml";;

module Leancop = struct

open Cop

let start opt lim =
  if !verbose then Format.printf "Start %d\n%!" (lim+1);
  let sub0 = Subst.empty 1000000 in
  let hashek = Mapping.fol_of_literal [] [] Hashek.hashek_tm in
  let hashek_neg = Fol_term.negate_lit hashek in
  let prfs = Search.prove_lit opt (sub0, 0) ([], [], lim) hashek_neg in
  match Strom.next_opt prfs with
    Some (sub, _, prf) -> Some (fst sub, prf)
  | None -> None

(* Possible optimization:
   Reduce variable indexes to reduce used substitution and fit better in cache *)
let LEANCOP_DEEPEN opt ths =
  Meson.clear_contrapos_cache(); Mapping.reset_vars(); Mapping.reset_consts();
  let ths' = ths @ Equality.create_eq_axs opt.add_eq_symmetry (map concl ths) in
  Database.axioms2db ths';
  let sub, prf = strategy opt start in
  let hashek = Mapping.fol_of_literal [] [] Hashek.hashek_tm in
  let hashek_neg = Fol_term.negate_lit hashek in
  let th = Recon.hashek_of_proof sub [] hashek_neg prf in
  EQ_MP Hashek.hashek_eq th

let PREPARE_LEAN_TAC ths =
  Prepare.PREPARE_COP_TAC ths THEN
  RULE_ASSUM_TAC(CONV_RULE(PRENEX_CONV THENC WEAK_CNF_CONV)) THEN
  RULE_ASSUM_TAC(repeat
   (fun th -> SPEC(genvar(type_of(fst(dest_forall(concl th))))) th)) THEN
  REPEAT (FIRST_X_ASSUM (Meson.CONJUNCTS_THEN' ASSUME_TAC)) THEN
  RULE_ASSUM_TAC(CONV_RULE(ASSOC_CONV DISJ_ASSOC)) THEN
  REPEAT (FIRST_X_ASSUM SUBST_VAR_TAC)

let PURE_LEANCOP opt gl =
  (FIRST_ASSUM CONTR_TAC ORELSE W(ACCEPT_TAC o LEANCOP_DEEPEN opt o map snd o fst)) gl

let GEN_LEANCOP_TAC opt ths = PREPARE_LEAN_TAC ths THEN PURE_LEANCOP opt
let LEANCOP_TAC ths = GEN_LEANCOP_TAC Cop.search_opts ths
let LEANCOP ths tm = prove(tm, LEANCOP_TAC ths)

end;;
