#load "unix.cma";;
#use "cop/leancop.ml";;

let time f x =
  if not (!report_timing) then f x else
  let start_time = Unix.gettimeofday () in
  try let result = f x in
      let finish_time = Unix.gettimeofday () in
      report("CPU time (user): "^(string_of_float(finish_time -. start_time)));
      result
  with e ->
      let finish_time = Unix.gettimeofday () in
      Format.print_string("Failed after (user) CPU time of "^
                          (string_of_float(finish_time -. start_time))^": ");
      raise e;;

meson_split_limit := 0;;

open Leancop;;

verbose := false;;

let LCCOMP_TAC ths =
  GEN_LEANCOP_TAC Cop.({search_opts with cut1 = false; cut2 = false; cut3 = false}) ths;;

let LCCOMP ths tm = prove(tm, LCCOMP_TAC ths);;

time (LCCOMP [APPEND]) `!h t. APPEND [(h : A)] t = CONS h t`;; (* 0.04s *)
time (MESON [APPEND]) `!h t. APPEND [(h : A)] t = CONS h t`;;   (* 0.05s *)

time (LCCOMP [LE; LT_NZ; NOT_LE; ONE; LT]) (concl LT);; (* 0.14s *)
(*
time (MESON [LE; LT_NZ; NOT_LE; ONE; LT]) (concl LT);;   (* Never terminates *)
*)

prioritize_real ();;
time (LCCOMP [real_lt; REAL_LE_TOTAL]) `!x y. x < y ==> x <= y`;; (* 0.03s *)
time (MESON [real_lt; REAL_LE_TOTAL]) `!x y. x < y ==> x <= y`;;   (* 0.03s *)

(* However: *)

time (LCCOMP [REAL_LE_LNEG; REAL_ADD_AC; real_lt]) (concl REAL_LT_RNEG);; (* 1.2s *)
time (MESON [REAL_LE_LNEG; REAL_ADD_AC; real_lt]) (concl REAL_LT_RNEG);;   (* 0.5s *)

