(* Load and process hashed problems. *)

needs "eval/timeout.ml";;
needs "eval/utils.ml";;

module Problem =
struct

let load_probs fname =
  Format.printf "Loading problems ...\n%!";
  let go probs line = match string_split_on_char ':' line with
      [t; ""] -> (t, []) :: probs
    | [t;  e] -> (t, string_split_on_char ' ' e) :: probs
    | _ -> failwith "read deps" in

  foldl_file go [] fname;;

let quote s = "\"" ^ s ^ "\""
let sqbr s = "[" ^ s ^ "]"
let print_problem c (conclh, premsh) conclth =
  let premshs = premsh |> List.map quote |> String.concat "; " |> sqbr in
  Printf.fprintf c "Hash: (%s, %s)\n%!" (quote conclh) premshs;
  Printf.fprintf c "To show: %s\n%!" (lstring_of_thm conclth)
(*
  print_endline "Premises:";
  List.iter (fun t -> print_thm t; print_newline ()) deps;
*)
;;

let find_prob (conclh, premsh) =
  let findn = find_thm o find_hash in
  try findn conclh, List.map findn premsh
  with Not_found -> failwith ("did not find prob " ^ conclh)

(* write theorem with a given problem hash to a new file in dir *)
let process_prob timeout tacstr (conclh, _ as probh) =
  Format.printf "Processing problem ...\n%!";

  let eval = try Sys.getenv "EVAL" with _ -> "HH" in
  let outdir = eval ^ "/toplevel-eval/" ^ string_of_int timeout ^ "s/" ^ tacstr in
  mkdir_p outdir 0o755;

  let oname = outdir ^ "/" ^ conclh ^ ".p" in
  Format.printf "Opening file %s ...\n%!" oname;
  let oc = open_out oname in
  Format.printf "File %s opened!\n%!" oname;

  let (conclth, prems) = find_prob probh in
  print_problem oc     probh conclth;
  print_problem stdout probh conclth;

  let run () =
    let tactic = Tactics.tactic_of_string tacstr in
    let th = prove (concl conclth, Tactics.run_tactic tactic prems) in
    Printf.fprintf oc "Theorem: %s\n%!" (lstring_of_thm th) in
  log_execution oc timeout run;
  close_out oc
;;

end
