#load "unix.cma";;

needs "eval/timeout.ml";;

(* Backported from OCaml 4.04 *)
let string_split_on_char sep s =
  let sub s ofs len =
    Bytes.unsafe_to_string (Bytes.sub (Bytes.unsafe_of_string s) ofs len) in
  let r = ref [] in
  let j = ref (String.length s) in
  for i = String.length s - 1 downto 0 do
    if String.get s i = sep then begin
      r := sub s (i + 1) (!j - i - 1) :: !r;
      j := i
    end
  done;
  sub s 0 !j :: !r

let mkdir_p dir perm =
  let go acc x =
    let acc' = acc ^ "/" ^ x in
    (try Unix.mkdir acc' perm with _ -> ());
    acc' in
  let acc0 = if Filename.is_relative dir then "." else "/" in
  ignore (List.fold_left go acc0 (string_split_on_char '/' dir))

(* left fold over every line of a file *)
let foldl_file f x file =
  let ic = Pervasives.open_in file
  and acc = ref x in
  (try while true do acc := f (!acc) (input_line ic) done
  with End_of_file -> close_in ic);
  !acc
;;

(* Unix.gettimeofday () under DMTCP 2.3.1 always yields the same time,
   therefore we parse the output of `date` *)
let timeofday () =
  try
    let ic = Unix.open_process_in "date +%s.%N" in
    let s = input_line ic in
    let time = float_of_string s in
    ignore (Unix.close_process_in ic);
    time
  with _ ->
    Format.printf "Reading time from `date` failed; using gettimeofday ()\n%!";
    Unix.gettimeofday ()
;;

let log_execution oc timeout f =
  try
    let t_start = timeofday () in
    let _ = with_timeout timeout f in
    let t_end = timeofday () in
    Printf.fprintf oc "Time taken: %.2f\n%!" (t_end -. t_start)
  with e ->
      Printf.fprintf oc "Exception: %s\n%!" (Printexc.to_string e);
      Printf.printf     "Exception: %s\n%!" (Printexc.to_string e)
